import polars as pl
from datetime import datetime
import re
from .models_definition import Definition
from .enums import FilterBinaryOperator, FilterBetweenOperator, FilterListOperator, FilterGroupOperator
from .models_query import FilterBinary, FilterBetween, FilterList, FilterGroup, FilterImpossible
from .filter_visitor import FilterVisitor


class FilterPolarsDataframeVisitor(FilterVisitor):
    def __init__(self, d: Definition, df: pl.DataFrame):
        super().__init__()
        self.d = d
        self.df = df
        self.property_dict = {p.key: p.alias for p in self.d.properties}

    def visit_binary(self, f: FilterBinary) -> any:
        match f.operator:
            case FilterBinaryOperator.equals:
                return pl.col(self.__get_alias(f.key)) == self.__get_filter_value(f.value)
            case FilterBinaryOperator.not_equals:
                return ~(pl.col(self.__get_alias(f.key)) == self.__get_filter_value(f.value))
            case FilterBinaryOperator.greater_than:
                return pl.col(self.__get_alias(f.key)) > self.__get_filter_value(f.value)
            case FilterBinaryOperator.greater_than_or_equals:
                return pl.col(self.__get_alias(f.key)) >= self.__get_filter_value(f.value)
            case FilterBinaryOperator.less_than:
                return pl.col(self.__get_alias(f.key)) < self.__get_filter_value(f.value)
            case FilterBinaryOperator.less_than_or_equals:
                return pl.col(self.__get_alias(f.key)) <= self.__get_filter_value(f.value)
            case FilterBinaryOperator.like:
                return pl.col(self.__get_alias(f.key)).str.contains(r"(?i)" + re.escape(f.value))
            case FilterBinaryOperator.not_like:
                return ~pl.col(self.__get_alias(f.key)).str.contains(r"(?i)" + re.escape(f.value))

    def visit_between(self, f: FilterBetween) -> any:
        match f.operator:
            case FilterBetweenOperator.between:
                return pl.col(self.__get_alias(f.key)).is_between(
                    self.__get_filter_value(f.value1),
                    self.__get_filter_value(f.value2))
            case FilterBetweenOperator.not_between:
                return ~pl.col(self.__get_alias(f.key)).is_between(
                    self.__get_filter_value(f.value1),
                    self.__get_filter_value(f.value2))

    def visit_list(self, f: FilterList) -> any:
        match f.operator:
            case FilterListOperator.value_in:
                return pl.col(self.__get_alias(f.key)).is_in(list(map(lambda x: self.__get_filter_value(x, True), f.values)))
            case FilterListOperator.value_not_in:
                return ~pl.col(self.__get_alias(f.key)).is_in(list(map(lambda x: self.__get_filter_value(x, True), f.values)))

    def visit_group(self, f: FilterGroup) -> any:
        r = None

        for i in f.filters:
            x = i.visit(self)

            if r is None:
                r = x
                continue

            if f.operator == FilterGroupOperator.group_and:
                r = (r & x)
            else:
                r = (r | x)

        return r

    def visit_impossible(self, f: FilterImpossible) -> any:
        return False

    def __get_alias(self, key: str) -> str:
        # TODO - die Funktion soll zentral zur Verfügung stehen und die Keys von den Properties und LoadOptions prüfen
        # Prüfen an div. anderen Stellen
        return self.property_dict[key]

    def __get_filter_value(self, value, ignore_string=False):
        if isinstance(value, datetime):
            return value
        if not ignore_string and isinstance(value, str):
            return pl.lit(value)

        return value



from dotenv import load_dotenv
from fastapi import FastAPI, Request, Response, HTTPException
import os
import uvicorn
from .caching import get_cached_items, run_scheduler
from .models_caching import Cache
from .models_definition import DefinitionCompact, Definition
from .query import LoadOptions
from .logger import init_logger
from .settings import get_settings
from . import query

app = FastAPI()


@app.get("/api/query", response_model_exclude_none=True)
def get_queries() -> list[DefinitionCompact]:
    return query.get_queries()


@app.get("/api/query/{key}", response_model_exclude_none=True)
def get_queries(key: str) -> Definition:
    definition = query.get_query(key)
    if definition is None:
        raise HTTPException(status_code=404, detail="Query nicht vorhanden")

    return definition


@app.post("/api/data", response_model=list[list])
def get_query_result(options: LoadOptions):
    r = query.get_query_result(options)
    return Response(content=r, media_type="application/json")


@app.get("/api/cache")
def get_cache_info() -> list[Cache]:
    cached_items = get_cached_items()

    return [Cache(
        key=i.cache_key,
        name=i.name,
        last_refresh=i.last_refresh,
        lazy=i.lazy,
        size_mb=i.get_file_size_mb(),
        is_invalidated=i.is_invalidated
    ) for i in cached_items]


@app.post("/api/cache/{key}/invalidate")
def invalid_cache(key: str):
    cached_items = get_cached_items()
    cached_item = next((i for i in cached_items if i.cache_key == key), None)

    if cached_item is None:
        raise HTTPException(status_code=404, detail="Cache nicht vorhanden")

    cached_item.invalidate()
    return Response(status_code=200)


def init():
    """
    Initialisiert die Umgebung
    :return: FastAPI
    """

    init_logger()
    load_dotenv()

    return app


def start():
    """
    Startet den Webserver
    :param api_key: Api-Key für das Abfragen der Definitionen und der Daten
    :return: None
    """

    api_key = os.getenv("API_KEY") or "TEST"

    @app.middleware("http")
    async def check_api_key(request: Request, call_next):
        if request.headers.get("X-Api-Key", "") != api_key:
            return Response(status_code=401, content="API Key ungültig")

        return await call_next(request)

    stop_event = run_scheduler()

    settings = get_settings()
    uvicorn.run(app, host=settings.host, port=settings.port, log_config={
        "version": 1,
        "disable_existing_loggers": False,
        "handlers": {}
    })

    stop_event.set()



from csv import excel
from pydantic import BaseModel, Field
from datetime import datetime
from typing import ForwardRef
from .enums import (
    PropertyAggregateType,
    FilterBinaryOperator,
    FilterBetweenOperator,
    FilterGroupOperator,
    FilterListOperator)
from .models_definition import Definition, DefinitionProperty

type RawValues = datetime | str | int | float | bool

# ForwardRef um zirkuläre Abhängigkeiten zu verhindern
FilterImpossibleT = ForwardRef("FilterImpossible")
FilterBinaryT = ForwardRef("FilterBinary")
FilterBetweenT = ForwardRef("FilterBetween")
FilterListT = ForwardRef("FilterList")
FilterGroupT = ForwardRef("FilterGroup")

type Filters = FilterBinaryT | FilterBetweenT | FilterListT | FilterGroupT | FilterImpossibleT

type Visitor = ForwardRef("odm_filter_visitor.Visitor")


class FilterBase(BaseModel):
    def visit(self, visitor: Visitor) -> any:
        pass


class FilterImpossible(FilterBase):
    impossible: bool

    def visit(self, visitor: Visitor) -> any:
        return visitor.visit_impossible(self)


class FilterBinary(FilterBase):
    key: str
    operator: FilterBinaryOperator
    value: RawValues

    def visit(self, visitor: Visitor) -> any:
        return visitor.visit_binary(self)


class FilterBetween(FilterBase):
    key: str
    operator: FilterBetweenOperator
    value1: RawValues
    value2: RawValues

    def visit(self, visitor: Visitor) -> any:
        return visitor.visit_between(self)


class FilterList(FilterBase):
    key: str
    operator: FilterListOperator
    values: list[RawValues]

    def visit(self, visitor: Visitor) -> any:
        return visitor.visit_list(self)


class FilterGroup(FilterBase):
    operator: FilterGroupOperator
    filters: list[Filters]

    def visit(self, visitor: Visitor) -> any:
        return visitor.visit_group(self)


class LoadProperty(BaseModel):
    key: str
    field_name: str = Field(serialization_alias="fieldName", validation_alias="fieldName")
    alias: str | None = Field(None, exclude=True)

    aggregate_type: PropertyAggregateType = Field(
        PropertyAggregateType.no,
        serialization_alias="aggregateType",
        validation_alias="aggregateType")


class LoadOptions(BaseModel):
    definition: Definition | None = Field(default=None, exclude=True)
    identity_key: str = Field(serialization_alias="identityKey", validation_alias="identityKey")

    key: str

    properties: list[LoadProperty]
    filter_blocks: list[str] = Field(serialization_alias="filterBlocks", validation_alias="filterBlocks")
    variables: list[Filters]

    filter: Filters | None
    relation_filter: Filters | None = Field(
        None,
        serialization_alias="relationFilter",
        validation_alias="relationFilter")
    max_rows: int | None = Field(None, serialization_alias="maxRows", validation_alias="maxRows")

    def get_definition_property(self, key: str) -> DefinitionProperty:
        return next(filter(lambda x: x.key == key, self.definition.properties), None)

    def get_variable_value(self, alias: str) -> FilterBase | None:
        v = next(filter(lambda x: x.alias == alias, self.definition.variables), None)
        if not v:
            return None

        return next(filter(lambda x: x.key == v.key, self.variables), None)

    def has_filter_block_by_alias(self, alias: str) -> bool:
        f = next(filter(lambda x: x.alias == alias, self.definition.filter_blocks), None)

        if f is None:
            return False

        return f.key in self.filter_blocks



import time
import uuid
import hashlib
from typing import Callable
from .logger import log_info


def create_uuid_from_string(val: str):
    """
    Konvertiert einen Text in eine reproduzierbare Guid
    :param val: Text
    :return: Guid als Text
    """

    hex_string = hashlib.md5(val.encode("UTF-8")).hexdigest()
    return str(uuid.UUID(hex=hex_string))


def timer(text: str, func: Callable):
    start = time.time()
    result = func()
    end = time.time()
    diff = round((end - start) * 1000, 0)

    log_info(f"{text}: {diff}ms")
    return result


import logging
import os
from logging.handlers import QueueHandler, QueueListener, TimedRotatingFileHandler
from queue import Queue

from .settings import get_settings

odm_logger = logging.getLogger("odm")


def init_logger():
    """
    Initialisiert den ODM-Logger
    :return: None
    """

    settings = get_settings()

    logger = logging.getLogger()
    logger.setLevel(logging.getLevelNamesMapping()[settings.logLevel.upper()])

    logger.handlers.clear()

    queue = Queue(-1)
    queue_handler = QueueHandler(queue)
    logger.addHandler(queue_handler)

    formatter = logging.Formatter("%(asctime)s [%(levelname)-5.5s]   %(message)s")

    console_handler = logging.StreamHandler()
    console_handler.setFormatter(formatter)

    if not os.path.exists("logs") or not os.path.isdir("logs"):
        os.makedirs("logs")

    file_handler = TimedRotatingFileHandler("logs/odm.log", when="midnight", interval=1, backupCount=7)
    file_handler.setFormatter(formatter)

    queue_listener = QueueListener(queue, console_handler, file_handler, respect_handler_level=True)
    queue_listener.start()

    __reset_sub_loggers()


def __reset_sub_loggers():
    for name in logging.root.manager.loggerDict:
        logger = logging.getLogger(name)
        logger.handlers.clear()
        logger.filters.clear()


def log_info(text):
    """
    Loggt eine Info-Nachricht
    :param text: Text
    :return: None
    """

    odm_logger.info(text)


def log_warning(text):
    """
    Loggt eine Warnung
    :param text: Text
    :return: None
    """

    odm_logger.warning(text)


def log_error(text):
    """
    Loggt einen Fehler
    :param text: Text
    :return: None
    """

    odm_logger.error(text)


def log_exception(text):
    """
    Loggt eine Exception
    :param text: Text
    :return: None
    """

    odm_logger.exception(text)


class InterceptHandler(logging.Handler):
    def emit(self, record: logging.LogRecord):
        odm_logger.log(record.levelno, record.getMessage())



import dataclasses
import datetime
import os.path
import threading
import time
from functools import wraps
from typing import Callable
import schedule
import polars as pl
from .logger import log_info, log_exception, log_error
from .utils import timer

type JobType = schedule.Job | list[schedule.Job] | str | list[str]
type DataFrame = pl.DataFrame | pl.LazyFrame


@dataclasses.dataclass
class CacheItem:
    cache_key: str
    name: str
    recreate: bool
    lazy: bool
    func: Callable[[], DataFrame]
    data: DataFrame | None = None

    is_creating: bool = False
    is_invalidated: bool = False
    last_refresh: datetime.datetime | None = None

    def create(self):
        if self.is_creating:
            log_info(f"Cache [{self.cache_key}] wird bereits erstellt")
            return

        self.is_creating = True

        try:
            log_info(f"Cache [{self.cache_key}] wird neu erstellt")
            text = f"Cache [{self.cache_key}] wurde neu erstellt"
            data: DataFrame = timer(text, self.func)

            if not os.path.exists("cache") or not os.path.isdir("cache"):
                os.makedirs("cache")

            if isinstance(data, pl.LazyFrame):
                # TODO wird aktuell noch nicht unterstützt. Daher über DataFrame lösen
                # data.sink_parquet(self.__get_file_path())

                data.collect().write_parquet(self.__get_file_path())
            else:
                data.write_parquet(self.__get_file_path())

            if self.lazy:
                data = pl.scan_parquet(self.__get_file_path())

            self.data = data
            self.is_invalidated = False
            self.last_refresh = datetime.datetime.now()

            return data
        finally:
            self.is_creating = False

    def get_or_create(self, wait_if_invalidated: bool = False):
        data = self.data

        if data is None or (self.is_invalidated and wait_if_invalidated):
            data = self.create()

        return data

    def invalidate(self, is_startup: bool = False):
        if self.is_invalidated:
            log_info(f"Cache [{self.cache_key}] ist bereits invalidiert")
            return

        self.is_invalidated = True

        if is_startup and self.__cache_exists():
            self.__load_cache()
            log_info(f"Cache [{self.cache_key}] aus Dateisystem eingelesen")
            self.last_refresh = datetime.datetime.fromtimestamp(os.path.getmtime(self.__get_file_path()))
            self.is_invalidated = False
            return

        if self.recreate:
            recreate_list.append(self)
            log_info(f"Cache [{self.cache_key}] zur Neuerstellung vorgemerkt")
        else:
            log_info(f"Cache [{self.cache_key}] invalidiert")
            self.data = None

    def get_file_size_mb(self):
        if not self.__cache_exists():
            return None

        return round(os.path.getsize(self.__get_file_path()) / 1024 / 1024, 2)

    def __get_file_path(self):
        return f"cache/{self.cache_key}.parquet"

    def __cache_exists(self):
        return os.path.exists(self.__get_file_path()) and os.path.isfile(self.__get_file_path())

    def __load_cache(self):
        if self.lazy:
            self.data = pl.scan_parquet(self.__get_file_path())
        else:
            self.data = pl.read_parquet(self.__get_file_path())


cache: dict[str, CacheItem] = {}
recreate_list: list[CacheItem] = []


def data_cache(cache_key: str, name: str, jobs: JobType = None, recreate: bool = True, lazy: bool = True):
    """
    Liefert die Daten aus dem Cache oder erstellt diesen neu
    :param cache_key: Key des Caches
    :param name: Name des Caches
    :param jobs: Uhrzeit, Jobs, Liste von Uhrzeit oder Liste von Jobs
    :param recreate: Angabe, ob nach Invalidieren des Cache der Cache direkt neu befüllt werden soll.
    :param lazy: Angabe, ob das Ergebnis als DataFrame oder LazyDataFrame zurückgegeben werden soll.
    Funktioniert nur mit Methoden ohne Parameter!
    :return: Ergebnis der dekorierten Methode
    """
    if jobs is None:
        jobs = ["02:00"]
    if isinstance(jobs, str):
        jobs = [jobs]
    if isinstance(jobs, schedule.Job):
        jobs = [jobs]
    if isinstance(jobs, list):
        jobs = [job if isinstance(job, schedule.Job) else schedule.every().day.at(job) for job in jobs]

    def decorator(func):
        cache_item = CacheItem(cache_key=cache_key, name=name, recreate=recreate, func=func, lazy=lazy, data=None)
        cache[cache_key] = cache_item

        @wraps(func)
        def wrapper():
            return cache_item.get_or_create()

        cache_item.invalidate(True)

        for job in jobs:
            job.do(cache_item.invalidate)

        return wrapper

    return decorator


def get_cached_items():
    """
    Liefert alle Cache
    :return: Cache
    """

    return cache.values()


def get_cached_data(cache_key: str, wait_if_invalidated: bool = False):
    """
    Liefert die Daten aus dem Cache oder erstellt diesen neu
    :param cache_key: Key des Caches
    :param wait_if_invalidated: falls der Cache-Eintrag invalidiert ist und, dann wird gewartet, bis der Cache
    aktualisiert wurde.
    :return: Ergebnis der dekorierten Methode
    """

    cache_item = cache.get(cache_key, None)
    if cache_item is None:
        log_error(f"Cache [{cache_key}] nicht gefunden")
        return None

    return cache_item.get_or_create(wait_if_invalidated)


def invalidate_cached_data(cache_key: str):
    """
    Invalidiert den Cache
    :param cache_key: Key des Caches
    :return: None
    """

    cache_item = cache.get(cache_key, None)
    if cache_item is None:
        log_error(f"Cache [{cache_key}] nicht gefunden")
        return

    cache_item.invalidate()


def run_scheduler():
    stop_event = threading.Event()

    class ScheduleThread(threading.Thread):
        @classmethod
        def run(cls):
            while not stop_event.is_set():
                schedule.run_pending()
                cls.recreate()
                time.sleep(5)

        @classmethod
        def recreate(cls):
            while len(recreate_list) > 0:
                cache_item = recreate_list.pop()

                try:
                    if cache_item.is_invalidated:
                        cache_item.create()
                except Exception:
                    log_exception(f"Cache [{cache_item.cache_key}] konnte nicht neu erstellt werden")

    schedule_thread = ScheduleThread()
    schedule_thread.start()

    return stop_event



from datetime import datetime
from .models_definition import DefinitionProperty, DefinitionVariable
from .models_query import FilterBinary, FilterBetween, FilterList, FilterGroup, FilterImpossible, LoadOptions


class FilterDatatypeVisitor:
    def __init__(self, options: LoadOptions):
        super().__init__()
        self.options = options

    def visit_binary(self, f: FilterBinary) -> any:
        if self.__is_datetime(f.key):
            f.value = self.__get_datetime(f.value)

    def visit_between(self, f: FilterBetween) -> any:
        if self.__is_datetime(f.key):
            f.value1 = self.__get_datetime(f.value1)
            f.value2 = self.__get_datetime(f.value2)

    def visit_list(self, f: FilterList) -> any:
        if self.__is_datetime(f.key):
            f.values = list(map(lambda x: self.__get_datetime(x), f.values))

    def visit_group(self, f: FilterGroup) -> any:
        for f in f.filters:
            f.visit(self)

    def visit_impossible(self, f: FilterImpossible) -> any:
        pass

    def __is_datetime(self, key: str) -> bool:
        p: DefinitionProperty | None = next(filter(lambda x: x.key == key, self.options.definition.properties), None)
        if p:
            if p.type == "datetime":
                return True
            return False

        v: DefinitionVariable | None = next(filter(lambda x: x.key == key, self.options.definition.variables), None)
        if v:
            if v.type == "datetime":
                return True
            return False

        return False

    def __get_datetime(self, value: str) -> datetime:
        return datetime.fromisoformat(value).replace(tzinfo=None)



from .models_definition import (
    Definition,
    DefinitionProperty,
    DefinitionVariable,
    DefinitionFilterBlock
)

from .models_query import (
    LoadOptions,
    LoadProperty,
    FilterBinary,
    FilterBetween,
    FilterList,
    FilterGroup,
    FilterImpossible
)

from .query_utils import (
    get_date_columns,
    convert_to_dataframe,
    filter_dataframe,
    aggregate_dataframe,
    convert_aggregate_type,
    filter_aggregate_convert,
    final_expr_dataframe,
    convert_dataframe_to_result
)

from .enums import (
    FilterOperator,
    FilterBinaryOperator,
    FilterBetweenOperator,
    FilterListOperator,
    FilterGroupOperator,
    DataType,
    PropertyAggregateType
)

from .query import (
    add_query,
    DefinitionCallback,
    LoadCallback
)

from .definition_importer import (
    add_yaml,
    auto_import_path
)

from .caching import (
    data_cache,
    get_cached_data,
    invalidate_cached_data
)

from .logger import (
    log_info,
    log_error,
    log_warning,
    log_exception
)

from .api import (
    init,
    start
)



from .models_query import FilterBinary, FilterBetween, FilterList, FilterGroup, FilterImpossible


class FilterVisitor:
    def visit_binary(self, f: FilterBinary) -> any:
        pass

    def visit_between(self, f: FilterBetween) -> any:
        pass

    def visit_list(self, f: FilterList) -> any:
        pass

    def visit_group(self, f: FilterGroup) -> any:
        pass

    def visit_impossible(self, f: FilterImpossible) -> any:
        pass
